#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#

"""
Lockable Resources CLI entrypoint
"""

import sys
import fnmatch
from functools import update_wrapper

import click

from . import VERSION
from .app import LockableResourceApp
from .config import ConfigLoader

DEFAULT_JENKINS_URL = "https://127.0.0.1"


class ClickStreamer:

    """
    Provides interfaces for output formating and input prompting functionality using click utilities.
    """

    # pylint: disable=no-self-use
    def __call__(self, msg, **kwargs):
        click.secho(msg, **kwargs)

    def prompt(self, *args, **kwargs):
        return click.prompt(*args, **kwargs)

    def confirm(self, *args, **kwargs):
        return click.confirm(*args, **kwargs)

    def info(self, msg, fg="cyan", **kwargs):
        click.secho(msg, fg=fg, **kwargs)

    def highlight(self, msg, fg="purple", **kwargs):
        click.secho(msg, fg=fg, **kwargs)

    def warn(self, msg, fg="yellow", **kwargs):
        click.secho(f"WARN: {msg}", fg=fg, **kwargs)

    def error(self, msg, fg="red", **kwargs):
        click.secho(f"ERROR: {msg}", fg=fg, **kwargs)


default_streamer = ClickStreamer()


class AppFactory:
    """
    Application Factory wrapper.

    This class provides a simple factory pattern to create the application only if needed.
    """

    def __init__(self, ctor, *args, **kwargs):
        """
        Instantiates a instance of AppFactory

        Args:
            ctor (callable): The constructor function to instanciate the application. This can be a function or a class.
            args: The positionnal arguments to provide to the constructor when instantiating.
            kwargs: The named arguments to provide to the constructor when instantiating.

        The factory provides the property `app` that holds the application instance. The application will be created
        when accessing the property the first time.
        """
        self._app = None
        self._ctor = ctor
        self._args = args
        self._kwargs = kwargs

    @property
    def app(self):
        if not self._app:
            self._app = self.create_app()
        return self._app

    def create_app(self):
        return self._ctor(*self._args, **self._kwargs)


def pass_app(f):
    """
    Decorator to only pass the application from factory.
    """

    def new_func(*args, **kwargs):
        ctx = click.get_current_context()
        obj = ctx.ensure_object(AppFactory)
        return f(obj.app, *args, **kwargs)

    return update_wrapper(new_func, f)


def click_completion_install_option(f):

    if "click_completion" in sys.modules:

        def completion_install( # pylint: disable=inconsistent-return-statements
            ctx, attr, value
        ):
            if not value or ctx.resilient_parsing:
                return value
            shell, path = click_completion.core.install()
            click.echo("%s completion installed in %s" % (shell, path))
            ctx.exit()

        # Add autocompletion auto install support if click autocomplete is installed
        click_completion.init()

        f = click.option(
            "--install-completion",
            is_flag=True,
            callback=completion_install,
            expose_value=False,
            help="Install shell auto completion",
        )(f)
    return f


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(VERSION)
    ctx.exit()


def make_regex_pattern(pattern, regex):
    if pattern and not regex:
        regex = fnmatch.translate(pattern)
    return regex


CONFIG_PATHS = (
    "/etc/lockable-resources.yml",
    "~/.lockable-resources.yml",
    ".lockable-resources.yml",
)
CONTEXT_SETTINGS = ConfigLoader(CONFIG_PATHS).click_settings


@click.group("resource", context_settings=CONTEXT_SETTINGS)
@click.option(
    "-V",
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
)
@click.option(
    "--jenkins-url",
    help="Jenkins URL",
    default=DEFAULT_JENKINS_URL,
    envvar=["JENKINS_URL", "JENKINS_HOST"],
    show_envvar=True,
)
@click.option(
    "--jenkins-user",
    "-u",
    help="Jenkins User used to connect to jenkins",
    envvar="JENKINS_USER",
    show_envvar=True,
    prompt=True,
)
@click.option(
    "--jenkins-token",
    "-T",
    help="Jenkins token used to authenticate to jenkins",
    envvar="JENKINS_TOKEN",
    prompt=True,
    hide_input=True,
)
@click.option(
    "--filter", "-f", "filter_expr", help="Global resources filter", show_default=True
)
@click.option(
    "--no-interactive",
    "interactive",
    is_flag=True,
    default=True,
    help="Disables interactive input prompts",
)
@click.pass_context
def cli(ctx, jenkins_url, jenkins_user, jenkins_token, filter_expr, interactive):
    ctx.interactive = interactive
    ctx.obj = AppFactory(
        LockableResourceApp.from_default,
        default_streamer,
        jenkins_url,
        jenkins_user,
        jenkins_token,
        filter_expr,
        interactive,
    )


@cli.command()
@click.argument("pattern", required=False)
@click.option("--regex", "-e", "regex", help="Regex filter pattern for resources")
@click.option("--label", "-L", "label", help="Regex filter pattern for labels")
@click.option(
    "--force",
    "-f",
    "force",
    is_flag=True,
    help="Force reserving another resource if user already owns any",
)
@pass_app
@click.pass_context
def reserve(ctx, app, pattern, regex, label, force):
    """Reserve a resource for current user.  If PATTERN is not provided, reserve the first free resource.
    If a resource is already reserved for the current user, the user will be prompted to confirm reservation,
    or use the flag `--force` to force reservation without prompt."""
    regex = make_regex_pattern(pattern, regex)
    if ctx.parent.interactive and not force:
        force = None
    app.reserve(regex, label, force)


@cli.command()
@click.argument("pattern", required=False)
@click.option("--regex", "-e", "regex", help="Regex filter pattern for resources")
@click.option("--label", "-L", "label", help="Regex filter pattern for labels")
@pass_app
def unreserve(app, pattern, regex, label):
    """Unreserve a resource from current user. If PATTERN is not provided, releases all owned resources."""
    regex = make_regex_pattern(pattern, regex)
    app.unreserve(regex, label)


@cli.command("list")
@click.argument("pattern", required=False)
@click.option(
    "--short-name", "-s", is_flag=True, help="Output to a short name if possible"
)
@click.option("--regex", "-e", "regex", help="Regex filter pattern for resources")
@click.option("--label", "-L", "label", help="Regex filter pattern for labels")
@click.option(
    "--state",
    "-S",
    "state",
    help="Regex filter pattern for state",
    type=click.Choice(["free", "reserved", "locked"]),
)
@pass_app
def list_(app, pattern, short_name, regex, label, state):
    """List all available resources matching PATTERN if set."""
    regex = make_regex_pattern(pattern, regex)
    app.list(regex, label, state, short_name)


@cli.command()
@click.argument("pattern", required=False)
@click.option("--regex", "-e", "regex", help="Regex filter pattern for resources")
@click.option("--label", "-L", "label", help="Regex filter pattern for labels")
@click.option(
    "--state",
    "-S",
    "state",
    help="Regex filter pattern for state",
    type=click.Choice(["free", "reserved", "locked"]),
)
@pass_app
def info(app, pattern, regex, label, state):
    """Show info of all available resources matching PATTERN if set."""
    regex = make_regex_pattern(pattern, regex)
    app.info(regex, label, state)


@cli.command()
@click.option("--user", "-u", help="User owning the boards (default to jenkins-user)")
@click.option(
    "--short-name", "-s", is_flag=True, help="Output to a short name if possible"
)
@click.option(
    "--count",
    "-c",
    type=int,
    default=None,
    help="Max number of owned resources to display",
)
@click.option(
    "--index",
    "-i",
    type=int,
    default=None,
    help="Position of owned resource in list to select",
)
@click.option(
    "--reserve",
    "-r",
    "auto_reserve",
    is_flag=True,
    help="Auto reserve a resource if none owned",
)
@click.option(
    "--reserve-pattern", "-p", help="Fnmatch pattern to match resources to reserve"
)
@click.option(
    "--reserve-regex", "-e", help="Regex pattern to match resources to reserve"
)
@click.option(
    "--reserve-label", "-L", "reserve_label", help="Regex filter pattern for labels"
)
@pass_app
def owned(
    app,
    user,
    short_name,
    count,
    index,
    auto_reserve,
    reserve_pattern,
    reserve_regex,
    reserve_label,
):
    """List owned resources by <user>."""
    reserve_regex = make_regex_pattern(reserve_pattern, reserve_regex)
    app.owned(
        user, short_name, count, index, auto_reserve, reserve_regex, reserve_label
    )


@cli.command()
@pass_app
def launch(app):
    """Lunches the web site of resources."""
    click.launch(app.obj.baseurl, wait=False)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#

"""
Jenkins Lockable Resources plugin API

This library aims to query Jenkins Lockable Resources plugin to retieve information or control resources.
"""

import re
import os
import shelve

from requests.exceptions import RequestException
from bs4 import BeautifulSoup
from jenkinsapi.jenkinsbase import JenkinsBase


def data_resource_from_row(row):
    name = str(row.attrs["data-resource-name"])
    columns = row.find_all("td")
    state_cell = columns[1]
    state_cell_text = state_cell.text.strip()
    parts = state_cell_text.split("\n")
    state, *parts = parts
    owner = parts[1].strip() if len(parts) > 1 else None
    label = str(columns[2].text)
    return dict(name=name, state=state, owner=owner, label=label)


class Resource:
    def __init__(
        self, jenkins_obj, name, state, owner, label=None, ephemeral=False, **kwargs
    ):
        self.jenkins = jenkins_obj.get_jenkins_obj()
        self.name = name
        self.state = state
        self.owner = owner
        self.label = label
        self.ephemeral = ephemeral
        self.baseurl = jenkins_obj.baseurl

    def is_locked(self):
        """
        Check if resource is locked
        """
        return self.state == "LOCKED"

    def is_reserved(self):
        """
        Check if resource is reserved
        """
        return self.state == "RESERVED"

    def is_free(self):
        """
        Check if resource is free
        """
        return self.state == "FREE"

    def is_owned(self):
        """
        Check if resource is owned by current user
        """
        return self.owner == self.jenkins.username

    def reserve(self):
        """
        Reserve the resource
        """
        url = f"{self.baseurl}/reserve?resource={self.name}"
        return self.jenkins.requester.get_url(url)

    def unreserve(self):
        """
        Unreserve the resource
        """
        url = f"{self.baseurl}/unreserve?resource={self.name}"
        return self.jenkins.requester.get_url(url)

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(
            dict(name=self.name, state=self.state, owner=self.owner, label=self.label)
        )


class LockableResources(JenkinsBase):

    """
    Class to hold information on locakble resources
    """

    def __init__(
        self,
        jenkins_obj,
        baseurl=None,
        poll=True,
        res_filter=None,
        use_cache=True,
        cache_file=None,
    ):
        """
        Init a lockable resource object

        Args:
            jenkins_obj (Jenkins): ref to the jenkins obj
            baseurl (str): basic url for querying information on a node
                If url is not set - object will construct it itself. This is
                useful when node is being created and not exists in Jenkins yet
            poll (bool): set to False if node does not exist or automatic
                refresh from Jenkins is not required. Default is True.
                If baseurl parameter is set to None - poll parameter will be
                set to False
            res_filter (str): Regex expression to filter resources (default '.*')
            use_cache (bool): Use persistant cache when server is unreachable
        """

        self.jenkins = jenkins_obj
        if not baseurl:
            poll = False
            baseurl = self.jenkins.baseurl + "/lockable-resources/"

        self._filter = re.compile(res_filter) if res_filter else re.compile(".*")
        self._cache_file = cache_file or os.path.expanduser("~/.lockable-resources.db")
        self._cache_data = shelve.open(self._cache_file) if use_cache else {}
        self._use_cache = use_cache
        self._data = None

        JenkinsBase.__init__(self, baseurl, poll=poll)

    def get_jenkins_obj(self):
        return self.jenkins

    def get_data(self, url, params=None, tree=None):
        """
        Retrieve data of lockable-resources
        """
        requester = self.get_jenkins_obj().requester
        try:
            response = requester.get_url(self.baseurl)
            if not response.ok:
                response.raise_for_status()

            # Scrap page for data
            html_soup = BeautifulSoup(response.text, "html.parser")
            main = html_soup.find("div", id="main-panel")
            table = main.find("table", class_="pane")
            rows = table.find_all("tr", attrs={"data-resource-name": self._filter})
            # Iterate through rows to extract resource name state and ownership
            self._cache_data["resources"] = list(map(data_resource_from_row, rows))
        except RequestException:
            if not self._use_cache or not self._cache_data:
                raise

        return self._cache_data

    @property
    def data(self):
        if self._data is None:
            self._data = self.poll()
        return self._data.get("resources", [])

    def list_resources(self) -> list:
        """
        Lists resources names
        """
        return list(self.keys())

    def get_resources(self) -> list:
        """
        Get resources
        """
        return list(self.values())

    def is_reserved(self, name) -> bool:
        """
        Check if a resource is reserved

        Args:
            name (str): The resource name
        """
        return self[name].is_reserved()

    def is_free(self, name) -> bool:
        """
        Check if resource is free
        """
        return self[name].is_free()

    def get_owner(self, name) -> str:
        """
        Get resource owner

        Args:
            name (str): The resource name
        """
        return self[name].owner

    def get_owned_resources(self, user=None) -> list:
        """
        Get resources owned by user

        Args:
            user (str): The owner name. Default to the current jenkins user.
        """
        if user is None:
            user = self.jenkins.username

        return [res for res in self.values() if res.is_reserved and res.owner == user]

    def get_free_resources(self) -> list:
        """
        Find a free resource
        """
        return [res for res in self.values() if res.is_free()]

    def reserve(self, name):
        """
        Reserve a resource
        """
        self[name].reserve()

    def unreserve(self, name):
        """
        Unreserve a resource
        """
        self[name].unreserve()

    def items(self):
        """
        Iterate over the names & objects for all resources
        """
        for resource in self.values():
            yield resource.name, resource

    def keys(self):
        """
        Iterate over the names of all available resources
        """
        for row in self.data:
            yield row["name"]

    def values(self, name=None, label=None, state=None):
        """
        Iterate over all available resources

        Args:
            name (str): a regex match string for resource name
            label (str): a regex match string for resource name
            state (str): the state to match (case insensitive)
        """
        data = self.data
        if name is not None:
            if isinstance(name, str):
                name = re.compile(name)
            data = filter(lambda x: name.match(x["name"]), data)
        if label is not None:
            if isinstance(label, str):
                label = re.compile(label)
            data = filter(lambda x: label.match(x["label"]), data)
        if state is not None:
            state = state.upper()
            data = filter(lambda x: x["state"] == state, data)

        for attrs in data:
            yield self._make_resource(attrs)

    def _make_resource(self, attrs):
        return Resource(self, **attrs)

    def __contains__(self, resource):
        """
        True if resource exists in Jenkins
        """
        return resource in self.keys()

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        # Regex key search
        pattern = re.compile(key)
        for attrs in self.data:
            if pattern.match(attrs["name"]):
                return self._make_resource(attrs)
        raise KeyError(key)

    def __str__(self):
        return self.baseurl



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Lockable Resources management application
"""
import logging
import functools

import requests.exceptions
from jenkinsapi.jenkins import Jenkins

from .model import LockableResources

INFO_STATE_COLORS = {"FREE": "green", "RESERVED": "yellow", "LOCKED": "red"}


def api_method(func):
    log = logging.getLogger(func.__name__)

    @functools.wraps(func)
    def wrapped(api, *args, **kwargs):
        try:
            return func(api, *args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            api.output.error(
                "Failed to connect to {}. Check your connection and try again.".format(
                    e.request.url
                )
            )
            log.error(e)
        except Exception as e:  # pylint: disable=broad-except
            api.output.error("{}".format(e))
            log.exception(e)

    return wrapped


class Application:
    """
    Base Application context class for accessing.
    """

    def __init__(self, output, obj, interactive=False):
        """
        Instanciate the application

        Args:
            output: The outputer object
            obj: A backend object
            interactive: Set interactive to let prompt for user input
        """
        self.output = output
        self.interactive = interactive
        self.obj = obj


class LockableResourceApp(Application):
    """
    Application context class for accessing lockable resources in jenkins.
    """

    @classmethod
    def from_default(
        cls,
        output,
        jenkins_url,
        jenkins_user,
        jenkins_token,
        filter_expr=None,
        interactive=False,
    ):
        """
        Default app creator using dependency injections
        - Instanciate Jenkins object
        - Instanciate LockableResources
        - Instanciate LockableResourceApp

        Args:
            output: The outputer object.
            jenkins_url: The url to jenkins instance
            jenkins_user: The user to authenticate to jenkins
            jenkins_token: The user token to use for authentication
            filter_expr: A filter expression to include only maching patterns of resources
            interactive: Set interactive to let prompt for user input

        Return:
            LockableResourceApp object
        """
        # Instanciate jenkins object
        jenkins = Jenkins(jenkins_url, jenkins_user, jenkins_token, lazy=True)
        mgr = LockableResources(jenkins, res_filter=filter_expr)

        return LockableResourceApp(output, mgr, interactive)

    @api_method
    def reserve(self, name=None, label=None, force=None):
        """
        Reserve a resource

        Args:
            name: Resource name to match (regex string)
            label: Resource label to match (regex string)
            force: Force reserve if already one or more resources owned
        """
        # Issue a warning if you already have one resource owned
        owned = self.obj.get_owned_resources()
        if owned:
            self.output.warn("You already have one resource owned.")
            if force is None:
                if not self.output.confirm("Force reserving?"):
                    return
            elif not force:
                return

        # Find free resources in list
        reslist = self._get_matching_free_resources(name, label)
        if reslist and not name:
            reslist = reslist[:1]

        if not reslist:
            self.output.warn("Sorry, no free resources at the moment. Try again later.")
            return

        for res in reslist:
            res.reserve()
            self.output.info(f"Reserved {res.name}")

    @api_method
    def unreserve(self, name=None, label=None):
        """
        Unreserve a resource

        Args:
            name: Resource name to match (regex string)
            label: Resource label to match (regex string)
        """
        # Name provided: Find matching resources
        reslist = self._get_matching_resources(name, label)

        # Find owned resources in list
        reslist = [r for r in reslist if r.is_owned()]

        if not reslist:
            self.output.warn("No resources to release")
            return

        for res in reslist:
            res.unreserve()
            self.output.info(f"Unreserved {res.name}")

    @api_method
    def list(self, name=None, label=None, state=None, short_name=False):
        """
        List resources

        Args:
            name: Resource name to match (regex string)
            label: Resource label to match (regex string)
            state: Match state given (case insensitive)
            short_name: Output to the short name version instead of full hostname
        """
        for res in self.obj.values(name, label, state):
            name = res.name
            if short_name:
                name = name.split(".")[0]
            self.output.info(name)

    @api_method
    def info(self, name=None, label=None, state=None):
        """
        Show info of resources

        Args:
            name: Resource name to match (regex string)
            label: Resource label to match (regex string)
            state: Match state given (case insensitive)
        """
        for res in self.obj.values(name, label, state):
            state = res.state
            if not res.is_free():
                state += f" by {res.owner}"
            self.output.info(f"{res.name}", nl=False)
            if res.label:
                self.output.info(f" [{res.label}]", nl=False)
            self.output.info(": ", nl=False)
            self.output.info(f"{state}", fg=INFO_STATE_COLORS[res.state])

    @api_method
    def owned(
        self,
        user=None,
        short_name=False,
        count=None,
        index=None,
        reserve=False,
        reserve_filter=None,
        reserve_label=None,
    ):
        """
        List owned resources

        Args:
            user: Owner of resource
            short_name: Output to the short name version instead of full hostname
            count: the max number of owned resources to return
            index: the owned resource list position to select
            reserve: try to reserve a resource matching the string or any if no resource currently owned
            reserve_filter: Resource pattern to include
            reserve_label: Resource label to match (regex string)
        """
        owned = self.obj.get_owned_resources(user)
        if not owned and reserve:

            free = self._get_matching_free_resources(reserve_filter, reserve_label)
            owned = free[:1]

            for res in owned:
                res.reserve()

        count = count or len(owned)
        if index is None:
            index = 0
        else:
            count = 1
        end = index + count

        for res in owned[index:end]:
            name = res.name
            if short_name:
                name = name.split(".")[0]
            self.output.info(name)

    def _get_matching_resources(self, name=r".*", label=None):
        # Find matchingn resources
        reslist = list(self.obj.values(name, label))
        if not reslist:
            exprs = []
            if name:
                exprs.append(f"name~=/{name}/")
            if label:
                exprs.append(f"label~=/{label}/")
            raise Exception("No resources matching " + " or ".join(exprs))
        return reslist

    def _get_matching_free_resources(self, name=r".*", label=None):
        # Find matchingn resources
        reslist = self._get_matching_resources(name, label)
        # Find free resources in list
        return [r for r in reslist if r.is_free()]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

from setuptools import setup

from jenkins_lockable_resources import VERSION


def get_version():
    CI_COMMIT_TAG = os.environ.get("CI_COMMIT_TAG", None)
    if not CI_COMMIT_TAG:
        return VERSION

    # Fixup extra release tag
    if VERSION not in CI_COMMIT_TAG:
        raise Exception(
            "Commit tag {} does not match version {}".format(CI_COMMIT_TAG, VERSION)
        )

    # Tag postfix
    postfix = CI_COMMIT_TAG.replace(VERSION, "")
    return VERSION + postfix


setup(version=get_version())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
CLI configuration utils
"""

import os
import collections

import yaml


class Config(dict):
    def __init__(self, *args, policy="overwrite"):
        self._policy = policy
        super().__init__(*args)

    def update(self, d):
        if not isinstance(d, dict):
            raise TypeError("Expected a dictionary")
        for k, v in d.items():
            if k not in self:
                self[k] = v
                continue

            # Key already exists in the dictionary
            merge_into = self[k]
            if isinstance(merge_into, collections.Mapping) and isinstance(
                v, collections.Mapping
            ):
                if not isinstance(merge_into, Config):
                    self[k] = Config(merge_into, policy=self._policy)
                self[k].update(v)
                continue

            if isinstance(merge_into, (list, tuple)) and isinstance(v, (list, tuple)):
                self[k] = v
                continue

            if not isinstance(v, collections.Mapping) and not isinstance(
                v, collections.Sequence
            ):
                self[k] = v
                continue

            if type(merge_into) is type(v):
                self[k] = v
                continue

            # Non matching merge types
            if self._policy == "strict":
                raise TypeError("Value type missmatch")
            if self._policy != "overwrite":
                continue

            self[k] = v


class ConfigLoader:
    """
    Loader class to load configuration file

    Config file holds tool parameters divided into sections or sub sections for each command or sub command:

    File structure:

        [/etc/myconf.yml]

            commamd_1:
                parameter_1: 'Test'
                parameter_2: 4

                sub_command:
                    parameter_1: true

            commamd_2:
                parameter_1: false

        [myconf.yml]

            commamd_2:
                parameter_1: true
        ...

    ConfigLoader can load a configuration dictionnary for the whole file or for a specific root section path.

    Example:
        This will load only the configuration for the command_1 section:

        >>> loader = ConfigLoader(config_files=['/etc/myconf.yml', 'myconf.yml'])
        >>> loader.conf
        {'commamd_1': {'parameter_1': 'Test', 'parameter_2': 4, 'sub_command': {'parameter_1': True}, 'commamd_2':
        'parameter_1': True}}

    """

    def __init__(self, config_files, defaults=None):
        """
        Create a new ConfigLoader instance

        Args:
            section (str, optional): the root section for the parser. Defaults to None.
            defaults (dict, optional): the default config values. Defaults to {}.
        """
        self._config_files = config_files
        self._defaults = defaults or {}
        self._conf = {}

    def _iter_conf_files(self):
        for path in self._config_files:
            path = os.path.expanduser(path)
            if os.path.exists(path):
                yield path

    def load(self):
        conf = Config(self._defaults.copy())
        for path in self._iter_conf_files():
            with open(path, "r") as f:
                c = yaml.safe_load(f)
            if c:
                conf.update(c)
        return conf

    @property
    def conf(self):
        if not self._conf:
            self._conf = self.load()
        return self._conf

    @property
    def click_settings(self):
        return dict(default_map=self.conf)
